from typing import Dict, Literal, Optional, Tuple
import chex
from jaxmarl.environments.multi_agent_env import MultiAgentEnv
from gymnax.environments import spaces
from brax import envs
import jax
import jax.numpy as jnp
from functools import partial

from .mappings import _agent_action_mapping, _agent_observation_mapping

# TODO: move homogenisation to a separate wrapper


class MABraxEnv(MultiAgentEnv):
    def __init__(
        self,
        env_name: str,
        episode_length: int = 1000,
        action_repeat: int = 1,
        auto_reset: bool = True,
        homogenisation_method: Optional[Literal["max", "concat"]] = None,
        backend: str = "positional",
        **kwargs
    ):
        """Multi-Agent Brax environment.

        Args:
            env_name: Name of the environment to be used.
            episode_length: Length of an episode. Defaults to 1000.
            action_repeat: How many repeated actions to take per environment
                step. Defaults to 1.
            auto_reset: Whether to automatically reset the environment when
                an episode ends. Defaults to True.
            homogenisation_method: Method to homogenise observations and actions
                across agents. If None, no homogenisation is performed, and
                observations and actions are returned as is. If "max", observations
                and actions are homogenised by taking the maximum dimension across
                all agents and zero-padding the rest. In this case, the index of the
                agent is prepended to the observation as a one-hot vector. If "concat",
                observations and actions are homogenised by masking the dimensions of
                the other agents with zeros in the full observation and action vectors.
                Defaults to None.
        """
        base_env_name = env_name.split("_")[0]
        env = envs.create(
            base_env_name, episode_length, action_repeat, auto_reset, backend=backend, **kwargs
        )
        self.env = env
        self.episode_length = episode_length
        self.action_repeat = action_repeat
        self.auto_reset = auto_reset
        self.homogenisation_method = homogenisation_method
        self.agent_obs_mapping = _agent_observation_mapping[env_name]
        self.agent_action_mapping = _agent_action_mapping[env_name]
        self.agents = list(self.agent_obs_mapping.keys())

        self.num_agents = len(self.agent_obs_mapping)
        obs_sizes = {
            agent: self.num_agents
            + max([o.size for o in self.agent_obs_mapping.values()])
            if homogenisation_method == "max"
            else self.env.observation_size
            if homogenisation_method == "concat"
            else obs.size
            for agent, obs in self.agent_obs_mapping.items()
        }
        act_sizes = {
            agent: max([a.size for a in self.agent_action_mapping.values()])
            if homogenisation_method == "max"
            else self.env.action_size
            if homogenisation_method == "concat"
            else act.size
            for agent, act in self.agent_action_mapping.items()
        }

        self.observation_spaces = {
            agent: spaces.Box(
                -jnp.inf,
                jnp.inf,
                shape=(obs_sizes[agent],),
            )
            for agent in self.agents
        }
        self.action_spaces = {
            agent: spaces.Box(
                -1.0,
                1.0,
                shape=(act_sizes[agent],),
            )
            for agent in self.agents
        }

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[Dict[str, chex.Array], envs.State]:
        state = self.env.reset(key)
        return self.get_obs(state), state

    @partial(jax.jit, static_argnums=(0,))
    def step_env(
        self,
        key: chex.PRNGKey,
        state: envs.State,
        actions: Dict[str, chex.Array],
    ) -> Tuple[
        Dict[str, chex.Array], envs.State, Dict[str, float], Dict[str, bool], Dict
    ]:
        global_action = self.map_agents_to_global_action(actions)
        next_state = self.env.step(state, global_action)  # type: ignore
        observations = self.get_obs(next_state)
        rewards = {agent: next_state.reward for agent in self.agents}
        rewards["__all__"] = next_state.reward
        dones = {agent: next_state.done.astype(jnp.bool_) for agent in self.agents}
        dones["__all__"] = next_state.done.astype(jnp.bool_)
        return (
            observations,
            next_state,  # type: ignore
            rewards,
            dones,
            next_state.info,
        )

    def get_obs(self, state: envs.State) -> Dict[str, chex.Array]:
        """Extracts agent observations from the global state.

        Args:
            state: Global state of the environment.

        Returns:
            A dictionary of observations for each agent.
        """
        return self.map_global_obs_to_agents(state.obs)

    def map_agents_to_global_action(
        self, agent_actions: Dict[str, jnp.ndarray]
    ) -> jnp.ndarray:
        global_action = jnp.zeros(self.env.action_size)
        for agent_name, action_indices in self.agent_action_mapping.items():
            if self.homogenisation_method == "max":
                global_action = global_action.at[action_indices].set(
                    agent_actions[agent_name][: action_indices.size]
                )
            elif self.homogenisation_method == "concat":
                global_action = global_action.at[action_indices].set(
                    agent_actions[agent_name][action_indices]
                )
            else:
                global_action = global_action.at[action_indices].set(
                    agent_actions[agent_name]
                )
        return global_action

    def map_global_obs_to_agents(
        self, global_obs: jnp.ndarray
    ) -> Dict[str, jnp.ndarray]:
        """Maps the global observation vector to the individual agent observations.

        Args:
            global_obs: The global observation vector.

        Returns:
            A dictionary mapping agent names to their observations. The mapping method
            is determined by the homogenisation_method parameter.
        """
        agent_obs = {}
        for agent_idx, (agent_name, obs_indices) in enumerate(
            self.agent_obs_mapping.items()
        ):
            if self.homogenisation_method == "max":
                # Vector with the agent idx one-hot encoded as the first num_agents
                # elements and then the agent's own observations (zero padded to
                # the size of the largest agent observation vector)
                agent_obs[agent_name] = (
                    jnp.zeros(
                        self.num_agents
                        + max([v.size for v in self.agent_obs_mapping.values()])
                    )
                    .at[agent_idx]
                    .set(1)
                    .at[agent_idx + 1 : agent_idx + 1 + obs_indices.size]
                    .set(global_obs[obs_indices])
                )
            elif self.homogenisation_method == "concat":
                # Zero vector except for the agent's own observations
                # (size of the global observation vector)
                agent_obs[agent_name] = (
                    jnp.zeros(global_obs.shape)
                    .at[obs_indices]
                    .set(global_obs[obs_indices])
                )
            else:
                # Just agent's own observations
                agent_obs[agent_name] = global_obs[obs_indices]
        return agent_obs

    @property
    def sys(self):
        return self.env.sys


class Ant(MABraxEnv):
    def __init__(self, **kwargs):
        super().__init__("ant_4x2", **kwargs)


class HalfCheetah(MABraxEnv):
    def __init__(self, **kwargs):
        super().__init__("halfcheetah_6x1", **kwargs)


class Hopper(MABraxEnv):
    def __init__(self, **kwargs):
        super().__init__("hopper_3x1", **kwargs)


class Humanoid(MABraxEnv):
    def __init__(self, **kwargs):
        super().__init__("humanoid_9|8", **kwargs)


class Walker2d(MABraxEnv):
    def __init__(self, **kwargs):
        super().__init__("walker2d_2x3", **kwargs)



from typing import Tuple, Callable

import chex
import jax.numpy as jnp
import jax.experimental.checkify as checkify
import jax.random

from .base import (
    ReplayBuffer, ReplayBufferState, Item, ItemBatch, IntScalar, ItemUpdateFn,
    make_default_add_batch_fn
)
from . import utils


@chex.dataclass(frozen=True)
class ClusteredReplayBufferState:
    cluster_buffer: ReplayBuffer
    cluster_state_batch: ReplayBufferState
    clustering_fn: Callable[[Item], IntScalar]
    distribution_power: float


def clustered_replay(
    num_clusters: int,
    cluster_buffer: ReplayBuffer,
    clustering_fn: Callable[[Item], IntScalar],
    distribution_power: float = 0.0,
) -> ReplayBuffer:
    def init_fn(item_prototype: Item) -> ClusteredReplayBufferState:
        cluster_states = [cluster_buffer.init_fn(item_prototype) for _ in range(num_clusters)]
        cluster_state_batch = utils.stack_trees(*cluster_states)
        return ClusteredReplayBufferState(
            cluster_buffer=cluster_buffer,
            cluster_state_batch=cluster_state_batch,
            clustering_fn=jax.tree_util.Partial(clustering_fn),
            distribution_power=distribution_power,
        )

    def size_fn(state: ClusteredReplayBufferState) -> IntScalar:
        return jnp.sum(jax.vmap(cluster_buffer.size_fn)(state.cluster_state_batch))

    def add_fn(state: ClusteredReplayBufferState, item: Item) -> ClusteredReplayBufferState:
        cluster_state_batch = state.cluster_state_batch
        cluster_index = state.clustering_fn(item)
        item_cluster = utils.get_pytree_batch_item(cluster_state_batch, cluster_index)
        item_cluster = state.cluster_buffer.add_fn(item_cluster, item)
        cluster_state_batch = utils.set_pytree_batch_item(cluster_state_batch, cluster_index, item_cluster)
        return state.replace(cluster_state_batch=cluster_state_batch)

    def sample_fn(state: ClusteredReplayBufferState, rng: chex.PRNGKey, batch_size: int) -> ItemBatch:
        cluster_sizes = jax.vmap(state.cluster_buffer.size_fn)(state.cluster_state_batch)
        cluster_weights = jnp.where(
            cluster_sizes > 0, jnp.power(cluster_sizes, state.distribution_power), cluster_sizes)

        cluster_fractions = cluster_weights / jnp.sum(cluster_weights)
        num_samples = jnp.round(batch_size * cluster_fractions).astype(jnp.int32)
        checkify.check(jnp.sum(num_samples) == batch_size, 'Number of samples does not match batch size')

        rng, cluster_selection_key = jax.random.split(rng)
        cluster_for_sample = jax.random.categorical(
            cluster_selection_key, logits=jnp.log(cluster_weights), shape=(batch_size,))
        rng_batch = jax.random.split(rng, batch_size)

        def sample_item(cluster_index: IntScalar, rng: chex.PRNGKey) -> Item:
            chex.assert_shape(cluster_index, ())

            cluster_state = utils.get_pytree_batch_item(state.cluster_state_batch, cluster_index)
            sample = state.cluster_buffer.sample_fn(cluster_state, rng, 1)

            chex.assert_tree_shape_prefix(sample, (1,))
            return utils.get_pytree_batch_item(sample, 0)

        return jax.vmap(sample_item)(cluster_for_sample, rng_batch)

    def update_fn(state: ClusteredReplayBufferState, item_update_fn: ItemUpdateFn) -> ClusteredReplayBufferState:
        def cluster_update_fn(cluster_state: ReplayBufferState) -> ReplayBufferState:
            return state.cluster_buffer.update_fn(cluster_state, item_update_fn)
        batch_cluster_update_fn = jax.vmap(cluster_update_fn)
        updated_cluster_state_batch = batch_cluster_update_fn(state.cluster_state_batch)
        return state.replace(cluster_state_batch=updated_cluster_state_batch)

    return ReplayBuffer(
        init_fn=jax.tree_util.Partial(init_fn),
        size_fn=jax.tree_util.Partial(size_fn),
        add_fn=jax.tree_util.Partial(add_fn),
        add_batch_fn=jax.tree_util.Partial(make_default_add_batch_fn(add_fn)),
        sample_fn=jax.tree_util.Partial(sample_fn),
        update_fn=jax.tree_util.Partial(update_fn),
    )



import math

import numpy as np

from jaxmarl.viz.window import Window
import jaxmarl.viz.grid_rendering as rendering
from jaxmarl.environments.overcooked.common import OBJECT_TO_INDEX, COLOR_TO_INDEX, COLORS


INDEX_TO_COLOR = [k for k,v in COLOR_TO_INDEX.items()]
TILE_PIXELS = 32

COLOR_TO_AGENT_INDEX = {0:0, 2:1} # Hardcoded. Red is first, blue is second

class OvercookedVisualizer:
	"""
	Manages a window and renders contents of EnvState instances to it.
	"""
	tile_cache = {}

	def __init__(self):
		self.window = None

	def _lazy_init_window(self):
		if self.window is None:
			self.window = Window('minimax')

	def show(self, block=False):
		self._lazy_init_window()
		self.window.show(block=block)

	def render(self, agent_view_size, state, highlight=True, tile_size=TILE_PIXELS):
		"""Method for rendering the state in a window. Esp. useful for interactive mode."""
		return self._render_state(agent_view_size, state, highlight, tile_size)

	def animate(self, state_seq, agent_view_size, filename="animation.gif"):
		"""Animate a gif give a state sequence and save if to file."""
		import imageio

		padding = agent_view_size - 2  # show

		def get_frame(state):
			grid = np.asarray(state.maze_map[padding:-padding, padding:-padding, :])
			# Render the state
			frame = OvercookedVisualizer._render_grid(
				grid,
				tile_size=TILE_PIXELS,
				highlight_mask=None,
				agent_dir_idx=state.agent_dir_idx,
				agent_inv=state.agent_inv
			)
			return frame

		frame_seq =[get_frame(state) for state in state_seq]

		imageio.mimsave(filename, frame_seq, 'GIF', duration=0.5)


	def render_grid(self, grid, tile_size=TILE_PIXELS, k_rot90=0, agent_dir_idx=None):
		self._lazy_init_window()

		img = OvercookedVisualizer._render_grid(
				grid,
				tile_size,
				highlight_mask=None,
				agent_dir_idx=agent_dir_idx,
			)
		# img = np.transpose(img, axes=(1,0,2))
		if k_rot90 > 0:
			img = np.rot90(img, k=k_rot90)

		self.window.show_img(img)

	def _render_state(self, agent_view_size, state, highlight=True, tile_size=TILE_PIXELS):
		"""
		Render the state
		"""
		self._lazy_init_window()

		padding = agent_view_size-2 # show
		grid = np.asarray(state.maze_map[padding:-padding, padding:-padding, :])
		grid_offset = np.array([1,1])
		h, w = grid.shape[:2]
		# === Compute highlight mask
		highlight_mask = np.zeros(shape=(h,w), dtype=bool)

		if highlight:
			# TODO: Fix this for multiple agents
			f_vec = state.agent_dir
			r_vec = np.array([-f_vec[1], f_vec[0]])

			fwd_bound1 = state.agent_pos
			fwd_bound2 = state.agent_pos + state.agent_dir*(agent_view_size-1)
			side_bound1 = state.agent_pos - r_vec*(agent_view_size//2)
			side_bound2 = state.agent_pos + r_vec*(agent_view_size//2)

			min_bound = np.min(np.stack([
						fwd_bound1,
						fwd_bound2,
						side_bound1,
						side_bound2]) + grid_offset, 0)

			min_y = min(max(min_bound[1], 0), highlight_mask.shape[0]-1)
			min_x = min(max(min_bound[0],0), highlight_mask.shape[1]-1)

			max_y = min(max(min_bound[1]+agent_view_size, 0), highlight_mask.shape[0]-1)
			max_x = min(max(min_bound[0]+agent_view_size, 0), highlight_mask.shape[1]-1)

			highlight_mask[min_y:max_y,min_x:max_x] = True

		# Render the whole grid
		img = OvercookedVisualizer._render_grid(
			grid,
			tile_size,
			highlight_mask=highlight_mask if highlight else None,
			agent_dir_idx=state.agent_dir_idx,
			agent_inv=state.agent_inv
		)
		self.window.show_img(img)

	@classmethod
	def _render_obj(
		cls,
		obj,
		img):
		# Render each kind of object
		obj_type = obj[0]
		color = INDEX_TO_COLOR[obj[1]]

		if obj_type == OBJECT_TO_INDEX['wall']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['goal']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])
			rendering.fill_coords(img, rendering.point_in_rect(0.1, 0.9, 0.1, 0.9), COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['agent']:
			agent_dir_idx = obj[2]
			tri_fn = rendering.point_in_triangle(
				(0.12, 0.19),
				(0.87, 0.50),
				(0.12, 0.81),
			)
			tri_fn = rendering.rotate_fn(tri_fn, cx=0.5, cy=0.5, theta=0.5*math.pi*agent_dir_idx)
			rendering.fill_coords(img, tri_fn, COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['empty']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['onion_pile']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])
			onion_fns = [rendering.point_in_circle(*coord, 0.15) for coord in [(0.5, 0.15), (0.3, 0.4), (0.8, 0.35),
																			  (0.4, 0.8), (0.75, 0.75)]]
			[rendering.fill_coords(img, onion_fn, COLORS[color]) for onion_fn in onion_fns]
		elif obj_type == OBJECT_TO_INDEX['onion']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])
			onion_fn = rendering.point_in_circle(0.5, 0.5, 0.15)
			rendering.fill_coords(img, onion_fn, COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['plate_pile']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])
			plate_fns = [rendering.point_in_circle(*coord, 0.2) for coord in [(0.3, 0.3), (0.75, 0.42),
																			  (0.4, 0.75)]]
			[rendering.fill_coords(img, plate_fn, COLORS[color]) for plate_fn in plate_fns]
		elif obj_type == OBJECT_TO_INDEX['plate']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])
			plate_fn = rendering.point_in_circle(0.5, 0.5, 0.2)
			rendering.fill_coords(img, plate_fn, COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['dish']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])
			plate_fn = rendering.point_in_circle(0.5, 0.5, 0.2)
			rendering.fill_coords(img, plate_fn, COLORS[color])
			onion_fn = rendering.point_in_circle(0.5, 0.5, 0.13)
			rendering.fill_coords(img, onion_fn, COLORS["orange"])
		elif obj_type == OBJECT_TO_INDEX['pot']:
			OvercookedVisualizer._render_pot(obj, img)
			# rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])
			# pot_fns = [rendering.point_in_rect(0.1, 0.9, 0.3, 0.9),
			# 		   rendering.point_in_rect(0.1, 0.9, 0.20, 0.23),
			# 		   rendering.point_in_rect(0.4, 0.6, 0.15, 0.20),]
			# [rendering.fill_coords(img, pot_fn, COLORS[color]) for pot_fn in pot_fns]
		else:
			raise ValueError(f'Rendering object at index {obj_type} is currently unsupported.')

	@classmethod
	def _render_pot(
		cls,
		obj,
		img):
		pot_status = obj[-1]
		num_onions = np.max([23-pot_status, 0])
		is_cooking = np.array((pot_status < 20) * (pot_status > 0))
		is_done = np.array(pot_status == 0)

		pot_fn = rendering.point_in_rect(0.1, 0.9, 0.33, 0.9)
		lid_fn = rendering.point_in_rect(0.1, 0.9, 0.21, 0.25)
		handle_fn = rendering.point_in_rect(0.4, 0.6, 0.16, 0.21)

		rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"])

		# Render onions in pot
		if num_onions > 0 and not is_done:
			onion_fns = [rendering.point_in_circle(*coord, 0.13) for coord in [(0.23, 0.33), (0.77, 0.33), (0.50, 0.33)]]
			onion_fns = onion_fns[:num_onions]
			[rendering.fill_coords(img, onion_fn, COLORS["yellow"]) for onion_fn in onion_fns]
			if not is_cooking:
				lid_fn = rendering.rotate_fn(lid_fn, cx=0.1, cy=0.25, theta=-0.1 * math.pi)
				handle_fn =rendering.rotate_fn(handle_fn, cx=0.1, cy=0.25, theta=-0.1 * math.pi)

		# Render done soup
		if is_done:
			soup_fn = rendering.point_in_rect(0.12, 0.88, 0.23, 0.35)
			rendering.fill_coords(img, soup_fn, COLORS["orange"])

		# Render the pot itself
		pot_fns = [pot_fn, lid_fn, handle_fn]
		[rendering.fill_coords(img, pot_fn, COLORS["black"]) for pot_fn in pot_fns]

		# Render progress bar
		if is_cooking:
			progress_fn = rendering.point_in_rect(0.1, 0.9-(0.9-0.1)/20*pot_status, 0.83, 0.88)
			rendering.fill_coords(img, progress_fn, COLORS["green"])


	@classmethod
	def _render_inv(
		cls,
		obj,
		img):
		# Render each kind of object
		obj_type = obj[0]
		if obj_type == OBJECT_TO_INDEX['empty']:
			pass
		elif obj_type == OBJECT_TO_INDEX['onion']:
			onion_fn = rendering.point_in_circle(0.75, 0.75, 0.15)
			rendering.fill_coords(img, onion_fn, COLORS["yellow"])
		elif obj_type == OBJECT_TO_INDEX['plate']:
			plate_fn = rendering.point_in_circle(0.75, 0.75, 0.2)
			rendering.fill_coords(img, plate_fn, COLORS["white"])
		elif obj_type == OBJECT_TO_INDEX['dish']:
			plate_fn = rendering.point_in_circle(0.75, 0.75, 0.2)
			rendering.fill_coords(img, plate_fn, COLORS["white"])
			onion_fn = rendering.point_in_circle(0.75, 0.75, 0.13)
			rendering.fill_coords(img, onion_fn, COLORS["orange"])
		else:
			raise ValueError(f'Rendering object at index {obj_type} is currently unsupported.')

	@classmethod
	def _render_tile(
		cls,
		obj,
		highlight=False,
		agent_dir_idx=None,
		agent_inv=None,
		tile_size=TILE_PIXELS,
		subdivs=3
	):
		"""
		Render a tile and cache the result
		"""
		# Hash map lookup key for the cache
		if obj is not None and obj[0] == OBJECT_TO_INDEX['agent']:
			# Get inventory of this specific agent
			if agent_inv is not None:
				color_idx = obj[1]
				agent_inv = agent_inv[COLOR_TO_AGENT_INDEX[color_idx]]
				agent_inv = np.array([agent_inv, -1, 0])

			if agent_dir_idx is not None:
				obj = np.array(obj)

				# TODO: Fix this for multiagents. Currently the orientation of other agents is wrong
				if len(agent_dir_idx) == 1:
					# Hacky way of making agent views orientations consistent with global view
					obj[-1] = agent_dir_idx[0]

		no_object = obj is None or (
			obj[0] in [OBJECT_TO_INDEX['empty'], OBJECT_TO_INDEX['unseen']] \
			and obj[2] == 0
		)

		if not no_object:
			if agent_inv is not None and obj[0] == OBJECT_TO_INDEX['agent']:
				key = (*obj, *agent_inv, highlight, tile_size)
			else:
				key = (*obj, highlight, tile_size)
		else:
			key = (obj, highlight, tile_size)

		if key in cls.tile_cache:
			return cls.tile_cache[key]

		img = np.zeros(shape=(tile_size * subdivs, tile_size * subdivs, 3), dtype=np.uint8)

		# Draw the grid lines (top and left edges)
		rendering.fill_coords(img, rendering.point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
		rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 0.031), (100, 100, 100))

		if not no_object:
			OvercookedVisualizer._render_obj(obj, img)
			# render inventory
			if agent_inv is not None and obj[0] == OBJECT_TO_INDEX['agent']:
				OvercookedVisualizer._render_inv(agent_inv, img)


		if highlight:
			rendering.highlight_img(img)

		# Downsample the image to perform supersampling/anti-aliasing
		img = rendering.downsample(img, subdivs)

		# Cache the rendered tile
		cls.tile_cache[key] = img

		return img

	@classmethod
	def _render_grid(
		cls,
		grid,
		tile_size=TILE_PIXELS,
		highlight_mask=None,
		agent_dir_idx=None,
		agent_inv=None):
		if highlight_mask is None:
			highlight_mask = np.zeros(shape=grid.shape[:2], dtype=bool)

		# Compute the total grid size in pixels
		width_px = grid.shape[1]*tile_size
		height_px = grid.shape[0]*tile_size

		img = np.zeros(shape=(height_px, width_px, 3), dtype=np.uint8)

		# Render the grid
		for y in range(grid.shape[0]):
			for x in range(grid.shape[1]):		
				obj = grid[y,x,:]
				if obj[0] in [OBJECT_TO_INDEX['empty'], OBJECT_TO_INDEX['unseen']] \
					and obj[2] == 0:
					obj = None

				tile_img = OvercookedVisualizer._render_tile(
					obj,
					highlight=highlight_mask[y, x],
					tile_size=tile_size,
					agent_dir_idx=agent_dir_idx,
					agent_inv=agent_inv,
				)

				ymin = y*tile_size
				ymax = (y+1)*tile_size
				xmin = x*tile_size
				xmax = (x+1)*tile_size
				img[ymin:ymax, xmin:xmax, :] = tile_img

		return img

	def close(self):
		self.window.close()


"""
End-to-End JAX Implementation of IQL.

Notice:
- Agents are controlled by a single RNN architecture.
- You can choose if sharing parameters between agents or not.
- Works also with non-homogenous agents (different obs/action spaces)
- Experience replay is a simple buffer with uniform sampling.
- Uses Double Q-Learning with a target agent network (hard-updated).
- You can use TD Loss (pymarl2) or DDQN loss (pymarl)
- Adam optimizer is used instead of RMSPROP.
- The environment is reset at the end of each episode.
- Trained with a team reward (reward['__all__'])
- At the moment, last_actions are not included in the agents' observations.

The implementation closely follows the original Pymarl: https://github.com/oxwhirl/pymarl/blob/master/src/learners/q_learner.py
"""

import os
import jax
import jax.numpy as jnp
import numpy as np

import optax
import flax.linen as nn
from flax.linen.initializers import constant, orthogonal
from flax.training.train_state import TrainState
import wandb
import hydra
from omegaconf import OmegaConf

from jaxmarl import make
from jaxmarl.environments.smax import map_name_to_scenario
from baselines.QLearning.utils import CTRolloutManager, EpsilonGreedy, Transition, UniformBuffer, ScannedRNN, save_params


class AgentRNN(nn.Module):
    # homogenous agent for parameters sharing, assumes all agents have same obs and action dim
    action_dim: int
    hidden_dim: int
    init_scale: float

    @nn.compact
    def __call__(self, hidden, x):
        obs, dones = x
        embedding = nn.Dense(self.hidden_dim, kernel_init=orthogonal(self.init_scale), bias_init=constant(0.0))(obs)
        embedding = nn.relu(embedding)

        rnn_in = (embedding, dones)
        hidden, embedding = ScannedRNN()(hidden, rnn_in)
        
        q_vals = nn.Dense(self.action_dim, kernel_init=orthogonal(self.init_scale), bias_init=constant(0.0))(embedding)

        return hidden, q_vals


def make_train(config, env):

    config["NUM_UPDATES"] = (
        config["TOTAL_TIMESTEPS"] // config["NUM_STEPS"] // config["NUM_ENVS"]
    )
    
    def train(rng):

        # INIT ENV
        rng, _rng = jax.random.split(rng)
        wrapped_env = CTRolloutManager(env, batch_size=config["NUM_ENVS"])
        test_env = CTRolloutManager(env, batch_size=config["NUM_TEST_EPISODES"]) # batched env for testing (has different batch size)
        init_obs, env_state = wrapped_env.batch_reset(_rng)
        init_dones = {agent:jnp.zeros((config["NUM_ENVS"]), dtype=bool) for agent in env.agents+['__all__']}

        # INIT BUFFER
        # to initalize the buffer is necessary to sample a trajectory to know its strucutre
        def _env_sample_step(env_state, unused):
            rng, key_a, key_s = jax.random.split(jax.random.PRNGKey(0), 3) # use a dummy rng here
            key_a = jax.random.split(key_a, env.num_agents)
            actions = {agent: wrapped_env.batch_sample(key_a[i], agent) for i, agent in enumerate(env.agents)}
            obs, env_state, rewards, dones, infos = wrapped_env.batch_step(key_s, env_state, actions)
            transition = Transition(obs, actions, rewards, dones)
            return env_state, transition
        _, sample_traj = jax.lax.scan(
            _env_sample_step, env_state, None, config["NUM_STEPS"]
        )
        sample_traj_unbatched = jax.tree_map(lambda x: x[:, 0], sample_traj) # remove the NUM_ENV dim
        buffer = UniformBuffer(parallel_envs=config["NUM_ENVS"], batch_size=config["BUFFER_BATCH_SIZE"], max_size=config["BUFFER_SIZE"])
        buffer_state = buffer.reset(sample_traj_unbatched)

        # INIT NETWORK
        agent = AgentRNN(action_dim=wrapped_env.max_action_space, hidden_dim=config["AGENT_HIDDEN_DIM"], init_scale=config['AGENT_INIT_SCALE'])
        rng, _rng = jax.random.split(rng)
        if config.get('PARAMETERS_SHARING', True):
            init_x = (
                jnp.zeros((1, 1, wrapped_env.obs_size)), # (time_step, batch_size, obs_size)
                jnp.zeros((1, 1)) # (time_step, batch size)
            )
            init_hs = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], 1) # (batch_size, hidden_dim)
            network_params = agent.init(_rng, init_hs, init_x)
        else:
            init_x = (
                jnp.zeros((len(env.agents), 1, 1, wrapped_env.obs_size)), # (time_step, batch_size, obs_size)
                jnp.zeros((len(env.agents), 1, 1)) # (time_step, batch size)
            )
            init_hs = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], len(env.agents),  1) # (n_agents, batch_size, hidden_dim)
            rngs = jax.random.split(_rng, len(env.agents)) # a random init for each agent
            network_params = jax.vmap(agent.init, in_axes=(0, 0, 0))(rngs, init_hs, init_x)

        # INIT TRAIN STATE AND OPTIMIZER
        def linear_schedule(count):
            frac = 1.0 - (count / config["NUM_UPDATES"])
            return config["LR"] * frac
        lr = linear_schedule if config.get('LR_LINEAR_DECAY', False) else config['LR']
        tx = optax.chain(
            optax.clip_by_global_norm(config["MAX_GRAD_NORM"]),
            optax.adam(learning_rate=lr, eps=config['EPS_ADAM'])
        )
        train_state = TrainState.create(
            apply_fn=agent.apply,
            params=network_params,
            tx=tx,
        )
        # target network params
        target_agent_params = jax.tree_map(lambda x: jnp.copy(x), train_state.params)

        # INIT EXPLORATION STRATEGY
        explorer = EpsilonGreedy(
            start_e=config["EPSILON_START"],
            end_e=config["EPSILON_FINISH"],
            duration=config["EPSILON_ANNEAL_TIME"]
        )

        # depending if using parameters sharing or not, q-values are computed using one or multiple parameters
        if config.get('PARAMETERS_SHARING', True):
            def homogeneous_pass(params, hidden_state, obs, dones):
                # concatenate agents and parallel envs to process them in one batch
                agents, flatten_agents_obs = zip(*obs.items())
                original_shape = flatten_agents_obs[0].shape # assumes obs shape is the same for all agents
                batched_input = (
                    jnp.concatenate(flatten_agents_obs, axis=1), # (time_step, n_agents*n_envs, obs_size)
                    jnp.concatenate([dones[agent] for agent in agents], axis=1), # ensure to not pass other keys (like __all__)
                )
                hidden_state, q_vals = agent.apply(params, hidden_state, batched_input)
                q_vals = q_vals.reshape(original_shape[0], len(agents), *original_shape[1:-1], -1) # (time_steps, n_agents, n_envs, action_dim)
                q_vals = {a:q_vals[:,i] for i,a in enumerate(agents)}
                return hidden_state, q_vals
        else:
            def homogeneous_pass(params, hidden_state, obs, dones):
                # homogeneous pass vmapped in respect to the agents parameters (i.e., no parameter sharing)
                agents, flatten_agents_obs = zip(*obs.items())
                batched_input = (
                    jnp.stack(flatten_agents_obs, axis=0), # (n_agents, time_step, n_envs, obs_size)
                    jnp.stack([dones[agent] for agent in agents], axis=0), # ensure to not pass other keys (like __all__)
                )
                # computes the q_vals with the params of each agent separately by vmapping
                hidden_state, q_vals = jax.vmap(agent.apply, in_axes=0)(params, hidden_state, batched_input)
                q_vals = {a:q_vals[i] for i,a in enumerate(agents)}
                return hidden_state, q_vals

        # TRAINING LOOP
        def _update_step(runner_state, unused):

            train_state, target_agent_params, env_state, buffer_state, time_state, init_obs, init_dones, test_metrics, rng = runner_state


            # EPISODE STEP
            def _env_step(step_state, unused):

                params, env_state, last_obs, last_dones, hstate, rng, t = step_state

                # prepare rngs for actions and step
                rng, key_a, key_s = jax.random.split(rng, 3)

                # SELECT ACTION
                # add a dummy time_step dimension to the agent input
                obs_   = {a:last_obs[a] for a in env.agents} # ensure to not pass the global state (obs["__all__"]) to the network
                obs_   = jax.tree_map(lambda x: x[np.newaxis, :], obs_)
                dones_ = jax.tree_map(lambda x: x[np.newaxis, :], last_dones)
                # get the q_values from the agent netwoek
                hstate, q_vals = homogeneous_pass(params, hstate, obs_, dones_)
                # remove the dummy time_step dimension and index qs by the valid actions of each agent 
                valid_q_vals = jax.tree_util.tree_map(lambda q, valid_idx: q.squeeze(0)[..., valid_idx], q_vals, wrapped_env.valid_actions)
                # explore with epsilon greedy_exploration
                actions = explorer.choose_actions(valid_q_vals, t, key_a)

                # STEP ENV
                obs, env_state, rewards, dones, infos = wrapped_env.batch_step(key_s, env_state, actions)
                transition = Transition(last_obs, actions, rewards, dones)

                step_state = (params, env_state, obs, dones, hstate, rng, t+1)
                return step_state, transition


            # prepare the step state and collect the episode trajectory
            rng, _rng = jax.random.split(rng)
            if config.get('PARAMETERS_SHARING', True):
                hstate = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], len(env.agents)*config["NUM_ENVS"]) # (n_agents*n_envs, hs_size)
            else:
                hstate = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], len(env.agents), config["NUM_ENVS"]) # (n_agents, n_envs, hs_size)

            step_state = (
                train_state.params,
                env_state,
                init_obs,
                init_dones,
                hstate, 
                _rng,
                time_state['timesteps'] # t is needed to compute epsilon
            )

            step_state, traj_batch = jax.lax.scan(
                _env_step, step_state, None, config["NUM_STEPS"]
            )

            # BUFFER UPDATE: save the collected trajectory in the buffer
            buffer_state = buffer.add(buffer_state, traj_batch)

            # LEARN PHASE
            def q_of_action(q, u):
                """index the q_values with action indices"""
                q_u = jnp.take_along_axis(q, jnp.expand_dims(u, axis=-1), axis=-1)
                return jnp.squeeze(q_u, axis=-1)

            def compute_target(target_max_qvals, rewards, dones):
                if config.get('TD_LAMBDA_LOSS', True):
                    # time difference loss
                    def _td_lambda_target(ret, values):
                        reward, done, target_qs = values
                        ret = jnp.where(
                            done,
                            target_qs,
                            ret*config['TD_LAMBDA']*config['GAMMA']
                            + reward
                            + (1-config['TD_LAMBDA'])*config['GAMMA']*(1-done)*target_qs
                        )
                        return ret, ret

                    ret = target_max_qvals[-1] * (1-dones[-1])
                    ret, td_targets = jax.lax.scan(
                        _td_lambda_target,
                        ret,
                        (rewards[-2::-1], dones[-2::-1], target_max_qvals[-1::-1])
                    )
                    targets = td_targets[::-1]
                else:
                    # standard DQN loss
                    targets = rewards[:-1] + config["GAMMA"]*(1-dones[:-1])*target_max_qvals
                return targets

            def _loss_fn(params, target_agent_params, init_hs, learn_traj):

                obs_ = {a:learn_traj.obs[a] for a in env.agents} # ensure to not pass the global state (obs["__all__"]) to the network
                _, q_vals = homogeneous_pass(params, init_hs, obs_, learn_traj.dones)
                _, target_q_vals = homogeneous_pass(target_agent_params, init_hs, obs_, learn_traj.dones)

                # get the q_vals of the taken actions (with exploration) for each agent
                chosen_action_qvals = jax.tree_map(
                    lambda q, u: q_of_action(q, u)[:-1], # avoid last timestep
                    q_vals,
                    learn_traj.actions
                )

                # get the target for each agent (assumes every agent has a reward)
                valid_q_vals = jax.tree_util.tree_map(lambda q, valid_idx: q[..., valid_idx], q_vals, wrapped_env.valid_actions)
                target_max_qvals = jax.tree_map(
                    lambda t_q, q: q_of_action(t_q, jnp.argmax(q, axis=-1))[1:], # avoid first timestep
                    target_q_vals,
                    jax.lax.stop_gradient(valid_q_vals)
                )

                # compute a single l2 loss for all the agents in one pass (parameter sharing)
                targets = jax.tree_map(
                    compute_target,
                    target_max_qvals,
                    {agent:learn_traj.rewards[agent] for agent in env.agents}, # rewards and agents could contain additional keys
                    {agent:learn_traj.dones[agent] for agent in env.agents}
                )
                chosen_action_qvals = jnp.concatenate(list(chosen_action_qvals.values()))
                targets = jnp.concatenate(list(targets.values()))

                if config.get('TD_LAMBDA_LOSS', True):
                    loss = jnp.mean(0.5*((chosen_action_qvals - jax.lax.stop_gradient(targets))**2))
                else:
                    loss = jnp.mean((chosen_action_qvals - jax.lax.stop_gradient(targets))**2)

                return loss


            # sample a batched trajectory from the buffer and set the time step dim in first axis
            rng, _rng = jax.random.split(rng)
            _, learn_traj = buffer.sample(buffer_state, _rng) # (batch_size, max_time_steps, ...)
            learn_traj = jax.tree_map(lambda x: jnp.swapaxes(x, 0, 1), learn_traj) # (max_time_steps, batch_size, ...)
            
            # for iql the loss must be computed differently with or without parameters sharing
            if config.get('PARAMETERS_SHARING', True):
                init_hs = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], len(env.agents)*config["BUFFER_BATCH_SIZE"]) # (n_agents*batch_size, hs_size)
                # compute loss and optimize grad
                grad_fn = jax.value_and_grad(_loss_fn, has_aux=False)
                loss, grads = grad_fn(train_state.params, target_agent_params, init_hs, learn_traj)
            else:
                # without parameters sharing, a different loss must be computed for each agent via vmap
                def _loss_fn(params, target_params, init_hs, obs, dones, actions, valid_actions, rewards):
                    _, q_vals = agent.apply(params, init_hs, (obs, dones))
                    _, target_q_vals = agent.apply(target_params, init_hs, (obs, dones))
                    chosen_action_qvals = q_of_action(q_vals, actions)[:-1]
                    valid_actions = valid_actions.reshape(*[1]*len(q_vals.shape[:-1]), -1) # reshape to match q_vals shape
                    valid_argmax = jnp.argmax(jnp.where(valid_actions.astype(bool), jax.lax.stop_gradient(q_vals), -1000000.), axis=-1)
                    target_max_qvals = q_of_action(target_q_vals, valid_argmax)[1:] # target q_vals of greedy actions, avoiding first time step
                    targets = compute_target(target_max_qvals, rewards, dones)
                    return jnp.mean((chosen_action_qvals - jax.lax.stop_gradient(targets))**2)
                init_hs = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], len(env.agents), config["BUFFER_BATCH_SIZE"]) # (n_agents, batch_size, hs_size)
                grad_fn = jax.value_and_grad(_loss_fn, has_aux=False)
                batchify = lambda x: jnp.stack([x[agent] for agent in env.agents], axis=0)
                loss, grads = jax.vmap(grad_fn, in_axes=0)(
                    train_state.params,
                    target_agent_params,
                    init_hs,
                    batchify(learn_traj.obs),
                    batchify(learn_traj.dones),
                    batchify(learn_traj.actions),
                    batchify(wrapped_env.valid_actions_oh),
                    batchify(learn_traj.rewards)
                )
                loss = loss.mean()

            # apply gradients
            train_state = train_state.apply_gradients(grads=grads)


            # UPDATE THE VARIABLES AND RETURN
            # reset the environment
            rng, _rng = jax.random.split(rng)
            init_obs, env_state = wrapped_env.batch_reset(_rng)
            init_dones = {agent:jnp.zeros((config["NUM_ENVS"]), dtype=bool) for agent in env.agents+['__all__']}

            # update the states
            time_state['timesteps'] = step_state[-1]
            time_state['updates']   = time_state['updates'] + 1

            # update the target network if necessary
            target_agent_params = jax.lax.cond(
                time_state['updates'] % config['TARGET_UPDATE_INTERVAL'] == 0,
                lambda _: jax.tree_map(lambda x: jnp.copy(x), train_state.params),
                lambda _: target_agent_params,
                operand=None
            )

            # update the greedy rewards
            rng, _rng = jax.random.split(rng)
            test_metrics = jax.lax.cond(
                time_state['updates'] % (config["TEST_INTERVAL"] // config["NUM_STEPS"] // config["NUM_ENVS"]) == 0,
                lambda _: get_greedy_metrics(_rng, train_state.params, time_state),
                lambda _: test_metrics,
                operand=None
            )

            # update the returning metrics
            metrics = {
                'timesteps': time_state['timesteps']*config['NUM_ENVS'],
                'updates' : time_state['updates'],
                'loss': loss,
                'rewards': jax.tree_util.tree_map(lambda x: jnp.sum(x, axis=0).mean(), traj_batch.rewards),
            }
            metrics.update(test_metrics) # add the test metrics dictionary

            if config.get('WANDB_ONLINE_REPORT', False):
                def callback(metrics):
                    wandb.log(
                        {
                            "returns": metrics['rewards']['__all__'].mean(),
                            "test_returns": metrics['test_returns']['__all__'].mean(),
                            "timestep": metrics['timesteps'],
                            "updates": metrics['updates'],
                            "loss": metrics['loss'],
                        }
                    )
                jax.debug.callback(callback, metrics)

            runner_state = (
                train_state,
                target_agent_params,
                env_state,
                buffer_state,
                time_state,
                init_obs,
                init_dones,
                test_metrics,
                rng
            )

            return runner_state, metrics

        def get_greedy_metrics(rng, params, time_state):
            """Help function to test greedy policy during training"""
            def _greedy_env_step(step_state, unused):
                params, env_state, last_obs, last_dones, hstate, rng = step_state
                rng, key_s = jax.random.split(rng)
                obs_   = {a:last_obs[a] for a in env.agents}
                obs_   = jax.tree_map(lambda x: x[np.newaxis, :], obs_)
                dones_ = jax.tree_map(lambda x: x[np.newaxis, :], last_dones)
                hstate, q_vals = homogeneous_pass(params, hstate, obs_, dones_)
                actions = jax.tree_util.tree_map(lambda q, valid_idx: jnp.argmax(q.squeeze(0)[..., valid_idx], axis=-1), q_vals, wrapped_env.valid_actions)
                obs, env_state, rewards, dones, infos = test_env.batch_step(key_s, env_state, actions)
                step_state = (params, env_state, obs, dones, hstate, rng)
                return step_state, (rewards, dones)
            rng, _rng = jax.random.split(rng)
            init_obs, env_state = test_env.batch_reset(_rng)
            init_dones = {agent:jnp.zeros((config["NUM_TEST_EPISODES"]), dtype=bool) for agent in env.agents+['__all__']}
            rng, _rng = jax.random.split(rng)
            if config["PARAMETERS_SHARING"]:
                hstate = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], len(env.agents)*config["NUM_TEST_EPISODES"]) # (n_agents*n_envs, hs_size)
            else:
                hstate = ScannedRNN.initialize_carry(config['AGENT_HIDDEN_DIM'], len(env.agents), config["NUM_TEST_EPISODES"]) # (n_agents, n_envs, hs_size)
            step_state = (
                params,
                env_state,
                init_obs,
                init_dones,
                hstate, 
                _rng,
            )
            step_state, rews_dones = jax.lax.scan(
                _greedy_env_step, step_state, None, config["NUM_STEPS"]
            )
            # compute the episode returns of the first episode that is done for each parallel env
            def first_episode_returns(rewards, dones):
                first_done = jax.lax.select(jnp.argmax(dones)==0., dones.size, jnp.argmax(dones))
                first_episode_mask = jnp.where(jnp.arange(dones.size) <= first_done, True, False)
                return jnp.where(first_episode_mask, rewards, 0.).sum()
            all_dones = rews_dones[1]['__all__']
            returns = jax.tree_map(lambda r: jax.vmap(first_episode_returns, in_axes=1)(r, all_dones), rews_dones[0])
            metrics = {
                'test_returns': returns # episode returns
            }
            if config.get('VERBOSE', False):
                def callback(timestep, val):
                    print(f"Timestep: {timestep}, return: {val}")
                jax.debug.callback(callback, time_state['timesteps']*config['NUM_ENVS'], returns['__all__'].mean())
            return metrics

        time_state = {
            'timesteps':jnp.array(0),
            'updates':  jnp.array(0)
        }
        rng, _rng = jax.random.split(rng)
        test_metrics = get_greedy_metrics(_rng, train_state.params, time_state) # initial greedy metrics
        
        # train
        rng, _rng = jax.random.split(rng)
        runner_state = (
            train_state,
            target_agent_params,
            env_state,
            buffer_state,
            time_state,
            init_obs,
            init_dones,
            test_metrics,
            _rng
        )
        runner_state, metrics = jax.lax.scan(
            _update_step, runner_state, None, config["NUM_UPDATES"]
        )
        return {'runner_state':runner_state, 'metrics':metrics}
    
    return train

@hydra.main(version_base=None, config_path="./config", config_name="config")
def main(config):
    config = OmegaConf.to_container(config)

    print('Config:\n', OmegaConf.to_yaml(config))

    env_name = config["env"]["ENV_NAME"]
    alg_name = f'iql_{"ps" if config["alg"].get("PARAMETERS_SHARING", True) else "ns"}'
    
    # smac init neeeds a scenario
    if 'SMAC' in env_name:
        config['env']['ENV_KWARGS']['scenario'] = map_name_to_scenario(config['env']['MAP_NAME'])
        env_name = 'jaxmarl_'+config['env']['MAP_NAME']

    env = make(config["env"]["ENV_NAME"], **config['env']['ENV_KWARGS'])
    config["alg"]["NUM_STEPS"] = config["alg"].get("NUM_STEPS", env.max_steps) # default steps defined by the env
    
    wandb.init(
        entity=config["ENTITY"],
        project=config["PROJECT"],
        tags=[
            alg_name.upper(),
            env_name.upper(),
            "RNN",
            "TD_LOSS" if config["alg"].get("TD_LAMBDA_LOSS", True) else "DQN_LOSS",
            f"jax_{jax.__version__}",
        ],
        name=f'{alg_name}_{env_name}',
        config=config,
        mode=config["WANDB_MODE"],
    )
    
    rng = jax.random.PRNGKey(config["SEED"])
    rngs = jax.random.split(rng, config["NUM_SEEDS"])
    train_vjit = jax.jit(jax.vmap(make_train(config["alg"], env)))
    outs = jax.block_until_ready(train_vjit(rngs))
    
    # save params
    if config['SAVE_PATH'] is not None:
        model_state = outs['runner_state'][0]
        params = jax.tree_map(lambda x: x[0], model_state.params) # save only params of the firt run
        save_dir = os.path.join(config['SAVE_PATH'], env_name)
        os.makedirs(save_dir, exist_ok=True)
        save_params(params, f'{save_dir}/{alg_name}.safetensors')
        print(f'Parameters of first batch saved in {save_dir}/{alg_name}.safetensors')


if __name__ == "__main__":
    main()
    


import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from flax import struct
from functools import partial
from jaxmarl.environments.mpe.simple import SimpleMPE, State
from jaxmarl.environments.mpe.default_params import *
from gymnax.environments.spaces import Box, Discrete

SPEAKER = "alice_0"
LISTENER = "bob_0"
ADVERSARY = "eve_0"
SPEAKER_IDX = 1
ADVERSARY_NAMES = ["eve_0"]
OBS_COLOUR = jnp.array([[255, 0, 0, 0], [0, 255, 0, 0]])


@struct.dataclass
class CryptoState(State):
    """State for the simple crypto environment."""

    goal_colour: chex.Array = None
    private_key: chex.Array = None


class SimpleCryptoMPE(SimpleMPE):
    """
    JAX Compatible version of simple_crypto_v2 PettingZoo environment.
    Source code: https://github.com/Farama-Foundation/PettingZoo/blob/master/pettingzoo/mpe/simple_crypto/simple_crypto.py
    Note, currently only have continuous actions implemented.
    """

    def __init__(self, num_agents=3, num_landmarks=2, action_type=DISCRETE_ACT):
        assert num_agents == 3, "Simple Crypto only supports 3 agents"
        assert num_landmarks == 2, "Simple Crypto only supports 2 landmarks"

        dim_c = 4  # Communication channel dimension

        num_landmarks = num_landmarks
        num_entities = num_landmarks + num_agents

        self.num_good_agents, self.num_adversaries = 2, 1
        self.num_agents = num_agents
        self.adversaries = [ADVERSARY]
        self.good_agents = [SPEAKER, LISTENER]

        assert self.num_agents == (self.num_good_agents + self.num_adversaries)
        assert len(self.adversaries) == self.num_adversaries
        assert len(self.good_agents) == self.num_good_agents

        agents = self.adversaries + self.good_agents
        assert agents[SPEAKER_IDX] == "alice_0"

        landmarks = ["landmark {}".format(i) for i in range(num_landmarks)]

        # Action and observation spaces
        if action_type == DISCRETE_ACT:
            action_spaces = {i: Discrete(4) for i in agents}
        elif action_type == CONTINUOUS_ACT:
            action_spaces = {i: Box(0.0, 1.0, (4,)) for i in agents}
        else:
            raise NotImplementedError("Action type not implemented")

        observation_spaces = {i: Box(-jnp.inf, jnp.inf, (4,)) for i in self.adversaries}
        observation_spaces.update(
            {i: Box(-jnp.inf, jnp.inf, (8,)) for i in self.good_agents}
        )

        colour = (
            [ADVERSARY_COLOUR] * self.num_adversaries
            + [AGENT_COLOUR] * self.num_good_agents
            + list(OBS_COLOUR)
        )

        # Parameters
        rad = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 0.05),
                jnp.full((self.num_good_agents), 0.05),
                jnp.full((num_landmarks), 0.05),
            ]
        )
        moveable = jnp.full((num_entities), False)
        silent = jnp.full((num_agents), 0)
        collide = jnp.full((num_entities), False)

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            action_spaces=action_spaces,
            observation_spaces=observation_spaces,
            dim_c=dim_c,
            colour=colour,
            rad=rad,
            moveable=moveable,
            silent=silent,
            collide=collide,
        )

    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, CryptoState]:
        key_a, key_l, key_g, key_k = jax.random.split(key, 4)

        p_pos = jnp.concatenate(
            [
                jax.random.uniform(key_a, (self.num_agents, 2), minval=-1, maxval=+1),
                jax.random.uniform(
                    key_l, (self.num_landmarks, 2), minval=-1.0, maxval=+1.0
                ),
            ]
        )

        g_idx = jax.random.randint(key_g, (1,), minval=0, maxval=self.num_landmarks)
        k_idx = jax.random.randint(key_k, (1,), minval=0, maxval=self.num_landmarks)

        state = CryptoState(
            p_pos=p_pos,
            p_vel=jnp.zeros((self.num_entities, self.dim_p)),
            c=jnp.zeros((self.num_agents, self.dim_c)),
            done=jnp.full((self.num_agents), False),
            step=0,
            goal_colour=jnp.array(
                OBS_COLOUR[g_idx], dtype=jnp.float32
            ).flatten(),  # set to float to be same as zoo env
            private_key=jnp.array(OBS_COLOUR[k_idx], dtype=jnp.float32).flatten(),
        )

        return self.get_obs(state), state

    @partial(jax.vmap, in_axes=[None, 0, 0])
    def _decode_continuous_action(
        self, a_idx: int, action: chex.Array
    ) -> Tuple[chex.Array, chex.Array]:
        """Communication action"""
        u = jnp.zeros((self.dim_p,))
        c = action
        return u, c

    @partial(jax.vmap, in_axes=[None, 0, 0])
    def _decode_discrete_action(
        self, a_idx: int, action: chex.Array
    ) -> Tuple[chex.Array, chex.Array]:
        """Communication action"""
        u = jnp.zeros((self.dim_p,))
        c = jnp.zeros((self.dim_c,))
        c = c.at[action].set(1.0)
        return u, c

    def get_obs(self, state: CryptoState) -> Dict[str, chex.Array]:
        goal_colour = state.goal_colour
        comm = state.c[SPEAKER_IDX]

        def _speaker():
            return jnp.concatenate(
                [
                    goal_colour,
                    state.private_key,
                ]
            )

        def _listener():
            return jnp.concatenate(
                [
                    state.private_key.flatten(),
                    comm,
                ]
            )

        def _adversary():
            return comm

        obs = {SPEAKER: _speaker(), LISTENER: _listener(), ADVERSARY: _adversary()}
        return obs

    def rewards(self, state: CryptoState) -> Dict[str, float]:
        comm_diff = jnp.sum(
            jnp.square(state.c - state.goal_colour), axis=1
        )  # check axis

        comm_zeros = ~jnp.all(state.c == 0)  # Ensure communication has happend

        mask = jnp.array([1, 0, -1])
        mask *= comm_zeros

        def _good():
            return jnp.sum(comm_diff * mask)

        def _adversary(idx):
            return -1 * jnp.sum(comm_diff[idx]) * comm_zeros

        rew = {a: _adversary(i) for i, a in enumerate(self.adversaries)}
        rew.update({a: _good() for i, a in enumerate(self.good_agents)})
        return rew



import jax.numpy as jnp
from flax.core.frozen_dict import FrozenDict

cramped_room = {
    "height" : 4,
    "width" : 5,
    "wall_idx" : jnp.array([0,1,2,3,4,
                            5,9,
                            10,14,
                            15,16,17,18,19]),
    "agent_idx" : jnp.array([6, 8]),
    "goal_idx" : jnp.array([18]),
    "plate_pile_idx" : jnp.array([16]),
    "onion_pile_idx" : jnp.array([5,9]),
    "pot_idx" : jnp.array([2])
}
asymm_advantages = {
    "height" : 5,
    "width" : 9,
    "wall_idx" : jnp.array([0,1,2,3,4,5,6,7,8,
                            9,11,12,13,14,15,17,
                            18,22,26,
                            27,31,35,
                            36,37,38,39,40,41,42,43,44]),
    "agent_idx" : jnp.array([29, 32]),
    "goal_idx" : jnp.array([12,17]),
    "plate_pile_idx" : jnp.array([39,41]),
    "onion_pile_idx" : jnp.array([9,14]),
    "pot_idx" : jnp.array([22,31])
}
coord_ring = {
    "height" : 5,
    "width" : 5,
    "wall_idx" : jnp.array([0,1,2,3,4,
                            5,9,
                            10,12,14,
                            15,19,
                            20,21,22,23,24]),
    "agent_idx" : jnp.array([7, 11]),
    "goal_idx" : jnp.array([22]),
    "plate_pile_idx" : jnp.array([10]),
    "onion_pile_idx" : jnp.array([15,21]),
    "pot_idx" : jnp.array([3,9])
}
forced_coord = {
    "height" : 5,
    "width" : 5,
    "wall_idx" : jnp.array([0,1,2,3,4,
                            5,7,9,
                            10,12,14,
                            15,17,19,
                            20,21,22,23,24]),
    "agent_idx" : jnp.array([11,8]),
    "goal_idx" : jnp.array([23]),
    "onion_pile_idx" : jnp.array([5,10]),
    "plate_pile_idx" : jnp.array([15]),
    "pot_idx" : jnp.array([3,9])
}

# Example of layout provided as a grid
counter_circuit_grid = """
WWWPPWWW
W A    W
B WWWW X
W     AW
WWWOOWWW
"""

def layout_grid_to_dict(grid):
    """Assumes `grid` is string representation of the layout, with 1 line per row, and the following symbols:
    W: wall
    A: agent
    X: goal
    B: plate (bowl) pile
    O: onion pile
    P: pot location
    ' ' (space) : empty cell
    """

    rows = grid.split('\n')

    if len(rows[0]) == 0:
        rows = rows[1:]
    if len(rows[-1]) == 0:
        rows = rows[:-1]

    keys = ["wall_idx", "agent_idx", "goal_idx", "plate_pile_idx", "onion_pile_idx", "pot_idx"]
    symbol_to_key = {"W" : "wall_idx",
                     "A" : "agent_idx",
                     "X" : "goal_idx",
                     "B" : "plate_pile_idx",
                     "O" : "onion_pile_idx",
                     "P" : "pot_idx"}

    layout_dict = {key : [] for key in keys}
    layout_dict["height"] = len(rows)
    layout_dict["width"] = len(rows[0])
    width = len(rows[0])

    for i, row in enumerate(rows):
        for j, obj in enumerate(row):
            idx = width * i + j
            if obj in symbol_to_key.keys():
                # Add object
                layout_dict[symbol_to_key[obj]].append(idx)
            if obj in ["X", "B", "O", "P"]:
                # These objects are also walls technically
                layout_dict["wall_idx"].append(idx)
            elif obj == " ":
                # Empty cell
                continue

    for key in symbol_to_key.values():
        # Transform lists to arrays
        layout_dict[key] = jnp.array(layout_dict[key])

    return FrozenDict(layout_dict)


overcooked_layouts = {
    "cramped_room" : FrozenDict(cramped_room),
    "asymm_advantages" : FrozenDict(asymm_advantages),
    "coord_ring" : FrozenDict(coord_ring),
    "forced_coord" : FrozenDict(forced_coord),
    "counter_circuit" : layout_grid_to_dict(counter_circuit_grid)
}


import jax
from jax import numpy as jnp
import numpy as np
import jax.experimental.checkify as checkify
import chex
import flax.linen as nn
from gymnax.environments.spaces import Box as BoxGymnax, Discrete as DiscreteGymnax
from jaxmarl.environments.spaces import Box, Discrete, MultiDiscrete
from jaxmarl.environments.multi_agent_env import MultiAgentEnv

from .buffers import uniform_replay
from .buffers.uniform import UniformReplayBufferState

from typing import NamedTuple, List
from functools import partial

from safetensors.flax import save_file, load_file
from flax.traverse_util import flatten_dict, unflatten_dict
import typing
import os 

def save_params(params: typing.Dict, filename: typing.Union[str, os.PathLike]) -> None:
    flattened_dict = flatten_dict(params, sep=',')
    save_file(flattened_dict, filename)

def load_params(filename: typing.Union[str, os.PathLike]) -> typing.Dict:
    flattened_dict = load_file(filename)
    return unflatten_dict(flattened_dict, sep=',')


def get_space_dim(space):
    if isinstance(space, (DiscreteGymnax, Discrete)):
        return space.n
    elif isinstance(space, (BoxGymnax, Box, MultiDiscrete)):
        return np.prod(space.shape)
    else:
        print(space)
        raise NotImplementedError('Current wrapper works only with Discrete/MultiDiscrete/Box action and obs spaces')


class CTRolloutManager:
    """
    Rollout Manager for Centralized Training with Parameters Sharing.
    - Batchify multiple environments (batch_size in __init__).
    - Adds a global state (obs["__all__"]) and a global reward (rewards["__all__"]) in the env.step returns.
    - Pads the observations of the agents in order to have all the same length.
    - Adds an agent id (one hot encoded) to the observation vectors.

    By default:
    - global_state is the concatenation of all agents' observations.
    - global_reward is the sum of all agents' rewards.
    """

    def __init__(self, env: MultiAgentEnv, batch_size:int, training_agents:List=None):
        self.env = env
        self.batch_size = batch_size

        self.agents = env.agents

        # the agents to train could differ from the total trainable agents in the env (f.i. if using pretrained agents)
        # it's important to know it in order to compute properly the default global rewards and state
        self.training_agents = self.agents if training_agents is None else training_agents    

        # TOREMOVE: this is because overcooked doesn't follow other envs conventions
        if len(env.observation_spaces) == 0:
            env.observation_spaces = {agent:env.observation_space() for agent in self.agents}
        if len(env.action_spaces) == 0:
            env.action_spaces = {agent:env.action_space() for agent in self.agents}
        
        # batched action sampling
        self.batch_samplers = {agent: jax.jit(jax.vmap(env.action_space(agent).sample, in_axes=0)) for agent in self.agents}

        # assumes the observations are flattened vectors
        self.max_obs_length = max(list(map(lambda x: get_space_dim(x), env.observation_spaces.values())))
        self.max_action_space = max(list(map(lambda x: get_space_dim(x), env.action_spaces.values())))
        self.obs_size = self.max_obs_length + len(self.agents)

        # agents ids
        self.agents_one_hot = {a:oh for a, oh in zip(self.agents, jnp.eye(len(self.agents)))}
        # valid actions
        self.valid_actions = {a:jnp.arange(u.n) for a, u in env.action_spaces.items()}
        self.valid_actions_oh ={a:jnp.concatenate((jnp.ones(u.n), jnp.zeros(self.max_action_space - u.n))) for a, u in env.action_spaces.items()}

        # custom global state and rewards for specific envs
        if 'smac' in env.name.lower():
            self.global_state = lambda obs, state: obs['world_state']
            self.global_reward = lambda rewards: rewards[self.training_agents[0]]
    
    @partial(jax.jit, static_argnums=0)
    def batch_reset(self, key):
        keys = jax.random.split(key, self.batch_size)
        return jax.vmap(self.wrapped_reset, in_axes=0)(keys)

    @partial(jax.jit, static_argnums=0)
    def batch_step(self, key, states, actions):
        keys = jax.random.split(key, self.batch_size)
        return jax.vmap(self.wrapped_step, in_axes=(0, 0, 0))(keys, states, actions)

    @partial(jax.jit, static_argnums=0)
    def wrapped_reset(self, key):
        obs_, state = self.env.reset(key)
        obs = jax.tree_util.tree_map(self._preprocess_obs, {agent:obs_[agent] for agent in self.agents}, self.agents_one_hot)
        obs["__all__"] = self.global_state(obs_, state)
        return obs, state

    @partial(jax.jit, static_argnums=0)
    def wrapped_step(self, key, state, actions):
        obs_, state, reward, done, infos = self.env.step(key, state, actions)
        obs = jax.tree_util.tree_map(self._preprocess_obs, {agent:obs_[agent] for agent in self.agents}, self.agents_one_hot)
        obs = jax.tree_util.tree_map(lambda d, o: jnp.where(d, 0., o), {agent:done[agent] for agent in self.agents}, obs) # ensure that the obs are 0s for done agents
        obs["__all__"] = self.global_state(obs_, state)
        reward["__all__"] = self.global_reward(reward)
        return obs, state, reward, done, infos

    @partial(jax.jit, static_argnums=0)
    def global_state(self, obs, state):
        return jnp.concatenate([obs[agent] for agent in self.agents], axis=-1)
    
    @partial(jax.jit, static_argnums=0)
    def global_reward(self, reward):
        return jnp.stack([reward[agent] for agent in self.training_agents]).sum(axis=0) 
    
    def batch_sample(self, key, agent):
        return self.batch_samplers[agent](jax.random.split(key, self.batch_size))

    @partial(jax.jit, static_argnums=0)
    def _preprocess_obs(self, arr, extra_features):
        # flatten
        arr = arr.flatten()
        # pad the observation vectors to the maximum length
        pad_width = [(0, 0)] * (arr.ndim - 1) + [(0, max(0, self.max_obs_length - arr.shape[-1]))]
        arr = jnp.pad(arr, pad_width, mode='constant', constant_values=0)
        # concatenate the extra features
        arr = jnp.concatenate((arr, extra_features), axis=-1)
        return arr


class EpsilonGreedy:
    """Epsilon Greedy action selection"""

    def __init__(self, start_e: float, end_e: float, duration: int):
        self.start_e  = start_e
        self.end_e    = end_e
        self.duration = duration
        self.slope    = (end_e - start_e) / duration
        
    @partial(jax.jit, static_argnums=0)
    def get_epsilon(self, t: int):
        e = self.slope*t + self.start_e
        return jnp.clip(e, self.end_e)
    
    @partial(jax.jit, static_argnums=0)
    def choose_actions(self, q_vals: dict, t: int, rng: chex.PRNGKey):
        
        def explore(q, eps, key):
            key_a, key_e   = jax.random.split(key, 2) # a key for sampling random actions and one for picking
            greedy_actions = jnp.argmax(q, axis=-1) # get the greedy actions 
            random_actions = jax.random.randint(key_a, shape=greedy_actions.shape, minval=0, maxval=q.shape[-1]) # sample random actions
            pick_random    = jax.random.uniform(key_e, greedy_actions.shape)<eps # pick which actions should be random
            chosed_actions = jnp.where(pick_random, random_actions, greedy_actions)
            return chosed_actions
        
        eps = self.get_epsilon(t)
        keys = dict(zip(q_vals.keys(), jax.random.split(rng, len(q_vals)))) # get a key for each agent
        choosed_actions = jax.tree_map(lambda q, k: explore(q, eps, k), q_vals, keys)
        return choosed_actions
    

class Transition(NamedTuple):
    obs: dict
    actions: dict
    rewards: dict
    dones: dict

class UniformBuffer:
    # Uniform Buffer replay buffer aggregating transitions from parallel envs
    # based on dejax: https://github.com/hr0nix/dejax/tree/main
    def __init__(self, parallel_envs:int=10, batch_size:int=32, max_size:int=5000):
        self.batch_size = batch_size
        self.buffer = uniform_replay(max_size=max_size)
        self.parallel_envs = parallel_envs
        self.sample = checkify.checkify(self.sample)
        
    def reset(self, transition_sample: Transition) -> UniformReplayBufferState:
        zero_transition = jax.tree_util.tree_map(jnp.zeros_like, transition_sample)
        return self.buffer.init_fn(zero_transition)
    
    @partial(jax.jit, static_argnums=0)
    def add(self, buffer_state: UniformReplayBufferState, transition: Transition) -> UniformReplayBufferState:
        def add_to_buffer(i, buffer_state):
            # assumes the transition is coming from jax.lax so the batch is on dimension 1
            return self.buffer.add_fn(buffer_state, jax.tree_util.tree_map(lambda x: x[:, i], transition))
        # need to use for and not vmap because you can't add multiple transitions on the same buffer in parallel
        return jax.lax.fori_loop(0, self.parallel_envs, add_to_buffer, buffer_state)
    
    @partial(jax.jit, static_argnums=0)
    def sample(self, buffer_state: UniformReplayBufferState, key: chex.PRNGKey) -> Transition:
        return self.buffer.sample_fn(buffer_state, key, self.batch_size)


class ScannedRNN(nn.Module):

    @partial(
        nn.scan,
        variable_broadcast="params",
        in_axes=0,
        out_axes=0,
        split_rngs={"params": False},
    )
    @nn.compact
    def __call__(self, carry, x):
        """Applies the module."""
        rnn_state = carry
        ins, resets = x
        hidden_size = ins.shape[-1]
        rnn_state = jnp.where(
            resets[:, np.newaxis],
            self.initialize_carry(hidden_size, *ins.shape[:-1]),
            rnn_state,
        )
        new_rnn_state, y = nn.GRUCell(hidden_size)(rnn_state, ins)
        return new_rnn_state, y

    @staticmethod
    def initialize_carry(hidden_size, *batch_size):
        # Use a dummy key since the default state init fn is just zeros.
        return nn.GRUCell(hidden_size, parent=None).initialize_carry(
            jax.random.PRNGKey(0), (*batch_size, hidden_size)
        )


from collections import OrderedDict
from enum import IntEnum

import numpy as np
import jax
import jax.numpy as jnp
from jax import lax
from jaxmarl.gridworld.env import Environment
from jaxmarl.environments import spaces
from typing import Tuple
import chex
from flax import struct

from jaxmarl.environments.overcooked.common import (
	OBJECT_TO_INDEX,
    COLOR_TO_INDEX,
	DIR_TO_VEC,
	make_maze_map)


class Actions(IntEnum):
    # Turn left, turn right, move forward
    left = 0
    right = 1
    forward = 2

    # Pick up an object
    pickup = 3
    # Drop an object
    drop = 4
    # Toggle/activate an object
    toggle = 5

    # Done completing task
    done = 6


@struct.dataclass
class EnvState:
	agent_pos: chex.Array
	agent_dir: chex.Array
	agent_dir_idx: int
	goal_pos: chex.Array
	wall_map: chex.Array
	maze_map: chex.Array
	time: int
	terminal: bool

@struct.dataclass
class EnvParams:
	height: int = 15
	width: int = 15
	n_walls: int = 25 
	agent_view_size: int = 5
	replace_wall_pos: bool = False
	see_through_walls: bool = True
	see_agent: bool = False
	normalize_obs: bool = False
	sample_n_walls: bool = False # Sample n_walls uniformly in [0, n_walls]
	max_episode_steps: int = 250
	singleton_seed: int = -1,


class Maze(Environment):
	def __init__(
		self,
		height=13,
		width=13,
		n_walls=25,
		agent_view_size=5,
		replace_wall_pos=False,
		see_agent=False,
		max_episode_steps=250,
		normalize_obs=False,
		singleton_seed=-1
	):
		super().__init__()

		self.obs_shape = (agent_view_size, agent_view_size, 3)
		self.action_set = jnp.array([
			Actions.left,
			Actions.right,
			Actions.forward,
			Actions.pickup,
			Actions.drop,
			Actions.toggle,
			Actions.done
		])

		self.params = EnvParams(
			height=height,
			width=width,
			n_walls=n_walls,
			agent_view_size=agent_view_size,
			replace_wall_pos=replace_wall_pos and not sample_n_walls,
			see_agent=see_agent,
			max_episode_steps=max_episode_steps,
			normalize_obs=normalize_obs,
			singleton_seed=-1,
		)

	@property
	def default_params(self) -> EnvParams:
		# Default environment parameters
		return EnvParams()

	def step_env(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: int,
	) -> Tuple[chex.Array, EnvState, float, bool, dict]:
		"""Perform single timestep state transition."""
		a = self.action_set[action]
		state, reward = self.step_agent(key, state, a)
		# Check game condition & no. steps for termination condition
		state = state.replace(time=state.time + 1)
		done = self.is_terminal(state)
		state = state.replace(terminal=done)

		return (
			lax.stop_gradient(self.get_obs(state)),
			lax.stop_gradient(state),
			reward.astype(jnp.float32),
			done,
			{},
		)

	def reset_env(
		self, 
		key: chex.PRNGKey, 
	) -> Tuple[chex.Array, EnvState]:
		"""Reset environment state by resampling contents of maze_map
		- initial agent position
		- goal position
		- wall positions
		"""
		params = self.params
		h = params.height
		w = params.width
		all_pos = np.arange(np.prod([h, w]), dtype=jnp.uint32)

		# Reset wall map, with shape H x W, and value of 1 at (i,j) iff there is a wall at (i,j)
		key, subkey = jax.random.split(key)
		wall_idx = jax.random.choice(
			subkey, all_pos, 
			shape=(params.n_walls,), 
			replace=params.replace_wall_pos)

		occupied_mask = jnp.zeros_like(all_pos)
		occupied_mask = occupied_mask.at[wall_idx].set(1)
		wall_map = occupied_mask.reshape(h, w).astype(jnp.bool_)

		# Reset agent position + dir
		key, subkey = jax.random.split(key)
		agent_idx = jax.random.choice(subkey, all_pos, shape=(1,), p=(~occupied_mask.astype(jnp.bool_)).astype(jnp.float32))
		occupied_mask = occupied_mask.at[agent_idx].set(1)
		agent_pos = jnp.array([agent_idx%w,agent_idx//w], dtype=jnp.uint32).flatten()

		key, subkey = jax.random.split(key)
		agent_dir_idx = jax.random.choice(subkey, jnp.arange(len(DIR_TO_VEC), dtype=jnp.uint8))
		agent_dir = DIR_TO_VEC.at[agent_dir_idx].get()

		# Reset goal position
		key, subkey = jax.random.split(key)
		goal_idx = jax.random.choice(subkey, all_pos, shape=(1,), p=(~occupied_mask.astype(jnp.bool_)).astype(jnp.float32))
		goal_pos = jnp.array([goal_idx%w,goal_idx//w], dtype=jnp.uint32).flatten()

		maze_map = make_maze_map(
			params,
			wall_map, 
			goal_pos, 
			agent_pos, 
			agent_dir_idx, 
			pad_obs=True)

		state = EnvState(
			agent_pos=agent_pos,
			agent_dir=agent_dir,
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=wall_map.astype(jnp.bool_),
			maze_map=maze_map,
			time=0,
			terminal=False,
		)

		return self.get_obs(state), state

	def get_obs(self, state: EnvState) -> chex.Array:
		"""Return limited grid view ahead of agent."""
		obs = jnp.zeros(self.obs_shape, dtype=jnp.uint8)
		
		agent_x, agent_y = state.agent_pos

		obs_fwd_bound1 = state.agent_pos
		obs_fwd_bound2 = state.agent_pos + state.agent_dir*(self.obs_shape[0]-1)

		side_offset = self.obs_shape[0]//2
		obs_side_bound1 = state.agent_pos + (state.agent_dir == 0)*side_offset
		obs_side_bound2 = state.agent_pos - (state.agent_dir == 0)*side_offset

		all_bounds = jnp.stack([obs_fwd_bound1, obs_fwd_bound2, obs_side_bound1, obs_side_bound2])

		# Clip obs to grid bounds appropriately
		padding = obs.shape[0]-1
		obs_bounds_min = np.min(all_bounds, 0) + padding
		obs_range_x = jnp.arange(obs.shape[0]) + obs_bounds_min[1]
		obs_range_y = jnp.arange(obs.shape[0]) + obs_bounds_min[0]

		meshgrid = jnp.meshgrid(obs_range_y,obs_range_x)
		coord_y = meshgrid[1].flatten()
		coord_x = meshgrid[0].flatten()

		obs = state.maze_map.at[
			coord_y,coord_x,:].get().reshape(obs.shape[0],obs.shape[1],3)

		obs = (state.agent_dir_idx == 0)*jnp.rot90(obs, 1) + \
			  	(state.agent_dir_idx == 1)*jnp.rot90(obs, 2) + \
			  	(state.agent_dir_idx == 2)*jnp.rot90(obs, 3) + \
			  	(state.agent_dir_idx == 3)*jnp.rot90(obs, 4)

		if not self.params.see_agent:
			obs = obs.at[-1, side_offset].set(
				jnp.array([OBJECT_TO_INDEX['empty'], 0, 0], dtype=jnp.uint8)
			)

		image = obs.astype(jnp.uint8)
		if self.params.normalize_obs:
			image = image/10.0

		obs_dict = dict(
			image=image,
			agent_dir=state.agent_dir_idx
		)

		return OrderedDict(obs_dict)

	def step_agent(self, key: chex.PRNGKey, state: EnvState, action: int) -> Tuple[EnvState, float]:
		params = self.params

		# Update agent position (forward action)
		fwd_pos = jnp.minimum(
			jnp.maximum(state.agent_pos + (action == Actions.forward)*state.agent_dir, 0), 
			jnp.array((params.width-1, params.height-1), dtype=jnp.uint32))

		# Can't go past wall or goal
		fwd_pos_has_wall = state.wall_map.at[fwd_pos[1], fwd_pos[0]].get()
		fwd_pos_has_goal = jnp.logical_and(fwd_pos[0] == state.goal_pos[0], fwd_pos[1] == state.goal_pos[1])

		fwd_pos_blocked = jnp.logical_or(fwd_pos_has_wall, fwd_pos_has_goal)

		agent_pos_prev = jnp.array(state.agent_pos)
		agent_pos = (fwd_pos_blocked*state.agent_pos + (~fwd_pos_blocked)*fwd_pos).astype(jnp.uint32)

		# Update agent direction (left_turn or right_turn action)
		agent_dir_offset = \
			0 \
			+ (action == Actions.left)*(-1) \
			+ (action == Actions.right)*1

		agent_dir_idx = (state.agent_dir_idx + agent_dir_offset) % 4
		agent_dir = DIR_TO_VEC[agent_dir_idx]

		# Update agent component in maze_map
		empty = jnp.array([OBJECT_TO_INDEX['empty'], 0, 0], dtype=jnp.uint8)
		agent = jnp.array([OBJECT_TO_INDEX['agent'], COLOR_TO_INDEX['red'], agent_dir_idx], dtype=jnp.uint8)
		padding = self.obs_shape[0]-1
		maze_map = state.maze_map
		maze_map = maze_map.at[padding+agent_pos_prev[1],padding+agent_pos_prev[0],:].set(empty)
		maze_map = maze_map.at[padding+agent_pos[1],padding+agent_pos[0],:].set(agent)

		reward = (1.0 - 0.9*((state.time+1)/params.max_episode_steps))*fwd_pos_has_goal

		return (
			state.replace(
				agent_pos=agent_pos,
				agent_dir_idx=agent_dir_idx,
				agent_dir=agent_dir,
				maze_map=maze_map,	
				terminal=fwd_pos_has_goal),
			reward
		)

	def is_terminal(self, state: EnvState) -> bool:
		"""Check whether state is terminal."""
		done_steps = state.time >= self.params.max_episode_steps
		return jnp.logical_or(done_steps, state.terminal)

	def get_eval_solved_rate_fn(self):
		def _fn(ep_stats):
			return ep_stats['return'] > 0

		return _fn

	@property
	def name(self) -> str:
		"""Environment name."""
		return "Maze"

	@property
	def num_actions(self) -> int:
		"""Number of actions possible in environment."""
		return len(self.action_set)

	def action_space(self) -> spaces.Discrete:
		"""Action space of the environment."""
		return spaces.Discrete(
			len(self.action_set),
			dtype=jnp.uint32
		)

	def observation_space(self) -> spaces.Dict:
		"""Observation space of the environment."""
		spaces_dict = {
			'image':spaces.Box(0, 255, self.obs_shape),
			'agent_dir': spaces.Discrete(4)
		}

		return spaces.Dict(spaces_dict)

	def state_space(self) -> spaces.Dict:
		"""State space of the environment."""
		params = self.params
		h = params.height
		w = params.width
		agent_view_size = params.agent_view_size
		return spaces.Dict({
			"agent_pos": spaces.Box(0, max(w, h), (2,), dtype=jnp.uint32),
			"agent_dir": spaces.Discrete(4),
			"goal_pos": spaces.Box(0, max(w, h), (2,), dtype=jnp.uint32),
			"maze_map": spaces.Box(0, 255, (w + agent_view_size, h + agent_view_size, 3), dtype=jnp.uint32),
			"time": spaces.Discrete(params.max_episode_steps),
			"terminal": spaces.Discrete(2),
		})

	def max_episode_steps(self) -> int:
		return self.params.max_episode_steps



from .environments import (
    SimpleMPE,
    SimpleTagMPE,
    SimpleWorldCommMPE,
    SimpleSpreadMPE,
    SimpleCryptoMPE,
    SimpleSpeakerListenerMPE,
    SimpleFacmacMPE,
    SimpleFacmacMPE3a,
    SimpleFacmacMPE6a,
    SimpleFacmacMPE9a,
    SimplePushMPE,
    SimpleAdversaryMPE,
    SimpleReferenceMPE,
    SMAX,
    HeuristicEnemySMAX,
    LearnedPolicyEnemySMAX,
    SwitchRiddle,
    Ant,
    Humanoid,
    Hopper,
    Walker2d,
    HalfCheetah,
    InTheGrid,
    InTheGrid_2p,
    HanabiGame,
    Overcooked,
    CoinGame,
)



def make(env_id: str, **env_kwargs):
    """A JAX-version of OpenAI's env.make(env_name), built off Gymnax"""
    if env_id not in registered_envs:
        raise ValueError(f"{env_id} is not in registered jaxmarl environments.")

    # 1. MPE PettingZoo Environments
    if env_id == "MPE_simple_v3":
        env = SimpleMPE(**env_kwargs)
    elif env_id == "MPE_simple_tag_v3":
        env = SimpleTagMPE(**env_kwargs)
    elif env_id == "MPE_simple_world_comm_v3":
        env = SimpleWorldCommMPE(**env_kwargs)
    elif env_id == "MPE_simple_spread_v3":
        env = SimpleSpreadMPE(**env_kwargs)
    elif env_id == "MPE_simple_crypto_v3":
        env = SimpleCryptoMPE(**env_kwargs)
    elif env_id == "MPE_simple_speaker_listener_v4":
        env = SimpleSpeakerListenerMPE(**env_kwargs)
    elif env_id == "MPE_simple_push_v3":
        env = SimplePushMPE(**env_kwargs)
    elif env_id == "MPE_simple_adversary_v3":
        env = SimpleAdversaryMPE(**env_kwargs)
    elif env_id == "MPE_simple_reference_v3":
        env = SimpleReferenceMPE(**env_kwargs)
    elif env_id == "MPE_simple_facmac_v1":
        env = SimpleFacmacMPE(**env_kwargs)
    elif env_id == "MPE_simple_facmac_3a_v1":
        env = SimpleFacmacMPE3a(**env_kwargs)
    elif env_id == "MPE_simple_facmac_6a_v1":
        env = SimpleFacmacMPE6a(**env_kwargs)
    elif env_id == "MPE_simple_facmac_9a_v1":
        env = SimpleFacmacMPE9a(**env_kwargs)

    # 2. Switch Riddle
    elif env_id == "switch_riddle":
        env = SwitchRiddle(**env_kwargs)

    # 3. SMAX
    elif env_id == "SMAX":
        env = SMAX(**env_kwargs)
    elif env_id == "HeuristicEnemySMAX":
        env = HeuristicEnemySMAX(**env_kwargs)
    elif env_id == "LearnedPolicyEnemySMAX":
        env = LearnedPolicyEnemySMAX(**env_kwargs)

    # 4. MABrax
    elif env_id == "ant_4x2":
        env = Ant(**env_kwargs)
    elif env_id == "halfcheetah_6x1":
        env = HalfCheetah(**env_kwargs)
    elif env_id == "hopper_3x1":
        env = Hopper(**env_kwargs)
    elif env_id == "humanoid_9|8":
        env = Humanoid(**env_kwargs)
    elif env_id == "walker2d_2x3":
        env = Walker2d(**env_kwargs)

    # 5. InTheGrid
    elif env_id == "storm":
        env = InTheGrid(**env_kwargs)
    # 5. InTheGrid
    elif env_id == "storm_2p":
        env = InTheGrid_2p(**env_kwargs)
    
    # 6. Hanabi
    elif env_id == "hanabi":
        env = HanabiGame(**env_kwargs)

    # 7. Overcooked
    elif env_id == "overcooked":
        env = Overcooked(**env_kwargs)

    # 8. Coin Game
    elif env_id == "coin_game":
        env = CoinGame(**env_kwargs)

    return env

registered_envs = [
    "MPE_simple_v3",
    "MPE_simple_tag_v3",
    "MPE_simple_world_comm_v3",
    "MPE_simple_spread_v3",
    "MPE_simple_crypto_v3",
    "MPE_simple_speaker_listener_v4",
    "MPE_simple_push_v3",
    "MPE_simple_adversary_v3",
    "MPE_simple_reference_v3",
    "MPE_simple_facmac_v1",
    "MPE_simple_facmac_3a_v1",
    "MPE_simple_facmac_6a_v1",
    "MPE_simple_facmac_9a_v1",
    "switch_riddle",
    "SMAX",
    "HeuristicEnemySMAX",
    "LearnedPolicyEnemySMAX",
    "ant_4x2",
    "halfcheetah_6x1",
    "hopper_3x1",
    "humanoid_9|8",
    "walker2d_2x3",
    "storm",
    "storm_2p",
    "hanabi",
    "overcooked",
    "coin_game",
]



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from functools import partial
from jaxmarl.environments.mpe.simple import SimpleMPE, State
from gymnax.environments.spaces import Box
from jaxmarl.environments.mpe.default_params import *


class SimpleTagMPE(SimpleMPE):
    def __init__(
        self,
        num_good_agents=1,
        num_adversaries=3,
        num_obs=2,
        action_type=DISCRETE_ACT,
    ):
        dim_c = 2  # NOTE follows code rather than docs

        num_agents = num_good_agents + num_adversaries
        num_landmarks = num_obs
        num_entities = num_agents + num_landmarks

        self.num_good_agents, self.num_adversaries = num_good_agents, num_adversaries

        self.adversaries = ["adversary_{}".format(i) for i in range(num_adversaries)]
        self.good_agents = ["agent_{}".format(i) for i in range(num_good_agents)]
        agents = self.adversaries + self.good_agents

        landmarks = ["landmark {}".format(i) for i in range(num_obs)]

        # Action and observation spaces
        observation_spaces = {
            i: Box(-jnp.inf, jnp.inf, (16,)) for i in self.adversaries
        }
        observation_spaces.update(
            {i: Box(-jnp.inf, jnp.inf, (14,)) for i in self.good_agents}
        )

        colour = (
            [ADVERSARY_COLOUR] * num_adversaries
            + [AGENT_COLOUR] * num_good_agents
            + [OBS_COLOUR] * num_obs
        )

        # Parameters
        rad = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 0.075),
                jnp.full((self.num_good_agents), 0.05),
                jnp.full((num_landmarks), 0.2),
            ]
        )
        accel = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 3.0),
                jnp.full((self.num_good_agents), 4.0),
            ]
        )
        max_speed = jnp.concatenate(
            [
                jnp.full((self.num_adversaries), 1.0),
                jnp.full((self.num_good_agents), 1.3),
                jnp.full((num_landmarks), 0.0),
            ]
        )
        collide = jnp.full((num_entities,), True)

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            observation_spaces=observation_spaces,
            dim_c=dim_c,
            colour=colour,
            rad=rad,
            accel=accel,
            max_speed=max_speed,
            collide=collide,
        )

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        @partial(jax.vmap, in_axes=(0))
        def _common_stats(aidx):
            """Values needed in all observations"""

            landmark_pos = (
                state.p_pos[self.num_agents :] - state.p_pos[aidx]
            )  # Landmark positions in agent reference frame

            # Zero out unseen agents with other_mask
            other_pos = state.p_pos[: self.num_agents] - state.p_pos[aidx]
            other_vel = state.p_vel[: self.num_agents]

            # use jnp.roll to remove ego agent from other_pos and other_vel arrays
            other_pos = jnp.roll(other_pos, shift=self.num_agents - aidx - 1, axis=0)[
                : self.num_agents - 1
            ]
            other_vel = jnp.roll(other_vel, shift=self.num_agents - aidx - 1, axis=0)[
                : self.num_agents - 1
            ]

            other_pos = jnp.roll(other_pos, shift=aidx, axis=0)
            other_vel = jnp.roll(other_vel, shift=aidx, axis=0)

            return landmark_pos, other_pos, other_vel

        landmark_pos, other_pos, other_vel = _common_stats(self.agent_range)

        def _good(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                ]
            )

        def _adversary(aidx):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                    other_vel[aidx, -1:].flatten(),  # 2
                ]
            )

        obs = {a: _adversary(i) for i, a in enumerate(self.adversaries)}
        obs.update(
            {a: _good(i + self.num_adversaries) for i, a in enumerate(self.good_agents)}
        )
        return obs

    def rewards(self, state: State) -> Dict[str, float]:
        @partial(jax.vmap, in_axes=(0, None))
        def _collisions(agent_idx: int, other_idx: int):
            return jax.vmap(self.is_collision, in_axes=(None, 0, None))(
                agent_idx,
                other_idx,
                state,
            )

        c = _collisions(
            jnp.arange(self.num_good_agents) + self.num_adversaries,
            jnp.arange(self.num_adversaries),
        )  # [agent, adversary, collison]

        def _good(aidx: int, collisions: chex.Array):
            rew = -10 * jnp.sum(collisions[aidx])

            mr = jnp.sum(self.map_bounds_reward(jnp.abs(state.p_pos[aidx])))
            rew -= mr
            return rew

        ad_rew = 10 * jnp.sum(c)

        rew = {a: ad_rew for a in self.adversaries}
        rew.update(
            {
                a: _good(i + self.num_adversaries, c)
                for i, a in enumerate(self.good_agents)
            }
        )
        return rew



import chex
import jax
import jax.numpy as jnp
from functools import partial


class Distribution:
    """Base class for all the start position and unit type generation"""

    def __init__(self, n_allies, n_enemies, map_width, map_height):
        self.n_allies = n_allies
        self.n_enemies = n_enemies
        self.map_width = map_width
        self.map_height = map_height

    @partial(jax.jit, static_argnums=(0,))
    def generate(self, key: chex.PRNGKey):
        raise NotImplementedError


class ReflectPositionDistribution(Distribution):
    @partial(jax.jit, static_argnums=(0,))
    def generate(self, key: chex.PRNGKey):
        key, ally_key = jax.random.split(key)
        ally_pos = jax.random.uniform(
            ally_key,
            shape=(self.n_allies, 2),
            minval=jnp.array([0.0, 0.0]),
            maxval=jnp.array([self.map_width / 2, self.map_height]),
        )
        enemy_pos = jnp.zeros((self.n_enemies, 2))
        min_size = min(self.n_enemies, self.n_allies)
        enemy_pos = enemy_pos.at[:min_size, 0].set(self.map_width - ally_pos[:, 0])
        enemy_pos = enemy_pos.at[:min_size, 1].set(ally_pos[:, 1])
        enemy_pos = jax.lax.select(
            min_size == self.n_enemies,
            enemy_pos,
            enemy_pos.at[self.n_allies :, :].set(
                jax.random.uniform(
                    key,
                    shape=(self.n_enemies - self.n_allies, 2),
                    minval=jnp.array([self.map_width / 2, 0.0]),
                    maxval=jnp.array([self.map_width, self.map_height]),
                )
            ),
        )
        return jnp.concatenate([ally_pos, enemy_pos])


class SurroundPositionDistribution(Distribution):
    @partial(jax.jit, static_argnums=(0,))
    def generate(self, key):
        # issue: want to randomly decide the centre and outside teams.
        # issue: don't know what that mean
        def draw_positions(key_, n_inside, n_outside):
            centre_pos = jnp.zeros((n_inside, 2))
            key_, centre_noise_key = jax.random.split(key_)
            centre_start_noise = jax.random.uniform(
                centre_noise_key, shape=(n_inside, 2), minval=-2, maxval=2
            )
            centre_pos = centre_pos.at[:, :].set(
                jnp.array([self.map_width / 2, self.map_height / 2])
                + centre_start_noise
            )
            n_groups = 4
            key_, key_groups = jax.random.split(key_)
            group_assignments = jax.random.categorical(
                key_groups,
                jnp.log(jnp.ones((n_groups,)) / n_groups),
                shape=(n_outside,),
            )
            centre = jnp.array(
                [[self.map_width / 2.0, self.map_height / 2.0]] * n_groups
            )
            edges = jnp.array(
                [
                    [0.0, 0.0],
                    [0.0, self.map_height],
                    [self.map_width, 0],
                    [self.map_width, self.map_height],
                ]
            )
            key_, t_key = jax.random.split(key_)
            t = jax.random.uniform(t_key, shape=(n_groups, 1), minval=0.0, maxval=1.0)
            group_positions = t * centre + (1 - t) * edges
            outside_pos = group_positions[group_assignments]
            key_, outside_noise_key = jax.random.split(key_)
            outside_pos_noise = jax.random.uniform(
                outside_noise_key, shape=(n_outside, 2), minval=-2, maxval=2
            )
            outside_pos = outside_pos + outside_pos_noise
            return {"outside": outside_pos, "inside": centre_pos}

        key, ally_key, enemy_key = jax.random.split(key, num=3)
        ally_inside_positions = draw_positions(ally_key, self.n_allies, self.n_enemies)
        ally_inside_positions = jnp.concatenate(
            [ally_inside_positions["inside"], ally_inside_positions["outside"]]
        )
        enemy_inside_positions = draw_positions(
            enemy_key, self.n_enemies, self.n_allies
        )
        enemy_inside_positions = jnp.concatenate(
            [enemy_inside_positions["outside"], enemy_inside_positions["inside"]]
        )
        ally_inside = jax.random.randint(key, shape=(), minval=0, maxval=2)
        return jax.lax.select(
            ally_inside, ally_inside_positions, enemy_inside_positions
        )


class SurroundAndReflectPositionDistribution(Distribution):
    def __init__(self, n_allies, n_enemies, map_width, map_height):
        super().__init__(n_allies, n_enemies, map_width, map_height)
        self.surround_distribution = SurroundPositionDistribution(
            n_allies, n_enemies, map_width, map_height
        )
        self.reflect_distribution = ReflectPositionDistribution(
            n_allies, n_enemies, map_width, map_height
        )

    def generate(self, key):
        key_draw, key_surround, key_reflect = jax.random.split(key, num=3)
        val = jax.random.uniform(key_draw)
        return jax.lax.select(
            val < 0.5,
            self.surround_distribution.generate(key_surround),
            self.reflect_distribution.generate(key_reflect),
        )


class UniformUnitTypeDistribution(Distribution):
    def __init__(self, n_allies, n_enemies, map_width, map_height, n_unit_types):
        super().__init__(n_allies, n_enemies, map_width, map_height)
        self.n_unit_types = n_unit_types

    def generate(self, key):
        enemy_key, ally_key = jax.random.split(key)
        ally_unit_types = jax.random.categorical(
            ally_key,
            jnp.log(jnp.ones((self.n_unit_types,)) / self.n_unit_types),
            shape=(self.n_allies,),
        ).astype(jnp.uint8)
        enemy_unit_types = jnp.zeros((self.n_enemies,), dtype=jnp.uint8)
        min_size = min(self.n_allies, self.n_enemies)
        enemy_unit_types = enemy_unit_types.at[:min_size].set(ally_unit_types)

        enemy_unit_types = jax.lax.select(
            min_size == self.n_enemies,
            enemy_unit_types,
            enemy_unit_types.at[min_size:].set(
                jax.random.categorical(
                    enemy_key,
                    jnp.log(jnp.ones((self.n_unit_types)) / self.n_unit_types),
                    shape=(self.n_enemies - self.n_allies,),
                ).astype(jnp.uint8)
            ),
        )
        return jnp.concatenate([ally_unit_types, enemy_unit_types], dtype=jnp.uint8)
